import click
import requests
import json
import textwrap
import time

try:
    import jq as pyjq
except ImportError:
    try:
        import pyjq
    except ImportError:
        pyjq = None


@click.command()
@click.version_option()
@click.argument("url", type=str, required=True)
@click.option("--nl", help="Output newline-delimited JSON", is_flag=True)
@click.option("--key", help="Top-level key to extract from each page")
@click.option("--jq", help="jq transformation to run on each page")
@click.option("--accept", help="Accept header to send")
@click.option("--sleep", help="Seconds to delay between requests", type=int)
@click.option("--silent", help="Don't show progress on stderr - default", is_flag=True)
@click.option("-v", "--verbose", help="Show progress on stderr", is_flag=True)
@click.option(
    "--show-headers", help="Dump response headers out to stderr", is_flag=True
)
@click.option(
    "--ignore-http-errors", help="Keep going on non-200 HTTP status codes", is_flag=True
)
@click.option(
    "--header", type=(str, str), multiple=True, help="Send custom request headers"
)
def cli(
    url,
    nl,
    key,
    jq,
    accept,
    sleep,
    silent,
    verbose,
    show_headers,
    ignore_http_errors,
    header,
):
    """
    Fetch paginated JSON from a URL

    Example usage:

    \b
        paginate-json https://api.github.com/repos/simonw/datasette/issues
    """
    # --silent is only in here for backwards compatibility
    silent = not verbose
    if jq and not pyjq:
        raise click.ClickException(
            "Missing dependency: 'pip install pyjq' for this to work"
        )
    if key and jq:
        raise click.ClickException("Can't use --key and --jq together")
    headers = {}
    for header_name, header_value in header:
        headers[header_name] = header_value
    if nl:
        for chunk in paginate(
            url=url,
            jq=jq,
            key=key,
            accept=accept,
            sleep=sleep,
            silent=silent,
            show_headers=show_headers,
            ignore_http_errors=ignore_http_errors,
            headers=headers,
        ):
            if not isinstance(chunk, list):
                chunk = [chunk]
            for row in chunk:
                click.echo(json.dumps(row))
    else:
        def iter_all():
            for chunk in paginate(
                url=url,
                jq=jq,
                key=key,
                accept=accept,
                sleep=sleep,
                silent=silent,
                show_headers=show_headers,
                ignore_http_errors=ignore_http_errors,
                headers=headers,
            ):
                if not isinstance(chunk, list):
                    chunk = [chunk]
                yield from chunk

        # Output JSON array by starting with '['
        # Then iterate two at a time to detect the last one
        first = True
        for item, last in enumerate_last(iter_all()):
            if first:
                # Output after first successful HTTP request, to
                # avoid outputting '[' if the first request fails
                click.echo("[")
                first = False
            click.echo(
                textwrap.indent(
                    json.dumps(item, indent=2).rstrip() + ("" if last else ","), "  "
                )
            )
        # Output closing ']'
        click.echo("]")


def paginate(
    *,
    url,
    jq=None,
    key=None,
    accept=None,
    sleep=None,
    silent=False,
    show_headers=False,
    ignore_http_errors=False,
    headers=None,
):
    while url:
        if not silent:
            click.echo(url, err=True)
        headers = headers or {}
        if accept is not None:
            headers["Accept"] = accept
        response = requests.get(url, headers=headers)
        if response.status_code != 200 and not ignore_http_errors:
            raise click.ClickException(
                "{} error fetching {}".format(response.status_code, url)
            )
        if show_headers:
            click.echo(
                json.dumps(dict(response.headers), indent=4, default=repr), err=True
            )
        try:
            next = response.links.get("next").get("url")
            # Resolve a potentially-relative URL to an absolute URL
            url = requests.compat.urljoin(url, next)
        except AttributeError:
            url = None
        if key:
            yield response.json()[key]
        elif jq:
            yield pyjq.first(jq, response.json())
        else:
            yield response.json()
        if sleep is not None:
            time.sleep(sleep)


def enumerate_last(iterable):
    it = iter(iterable)
    try:
        last = next(it)
    except StopIteration:
        return
    while True:
        try:
            current = next(it)
            yield last, False
            last = current
        except StopIteration:
            yield last, True
            break



from paginate_json import cli
from click.testing import CliRunner
import pytest


@pytest.mark.skipif(cli.pyjq, reason="pyjq is installed")
def test_pyjq_error():
    result = CliRunner().invoke(cli.cli, ["--jq=.", "http://example.com/"])
    assert result.exit_code != 0
    assert (
        result.output.strip()
        == "Error: Missing dependency: 'pip install pyjq' for this to work"
    )


@pytest.mark.parametrize("relative_urls", (False, True))
def test_paginate_json(requests_mock, relative_urls):
    requests_mock.get(
        "https://example.com/",
        json=[{"id": 1}, {"id": 2}],
        headers={
            "link": '<{}>; rel="next"'.format(
                "/?page=2" if relative_urls else "https://example.com/?page=2"
            )
        },
    )
    requests_mock.get("https://example.com/?page=2", json=[{"id": 3}, {"id": 4}])
    result = CliRunner(mix_stderr=False).invoke(cli.cli, ["https://example.com/"])
    assert result.exit_code == 0
    assert result.stdout == (
        '[\n  {\n    "id": 1\n  },\n  {\n    "id": 2\n  },\n'
        '  {\n    "id": 3\n  },\n  {\n    "id": 4\n  }\n]\n'
    )


def test_paginate_json_key(requests_mock):
    requests_mock.get(
        "https://example.com/",
        json={"rows": [1, 2]},
        headers={"link": '</?page=2>; rel="next"'},
    )
    requests_mock.get("https://example.com/?page=2", json={"rows": [3, 4]})
    result = CliRunner(mix_stderr=False).invoke(
        cli.cli, ["https://example.com/", "--key", "rows"]
    )
    assert result.exit_code == 0
    assert result.stdout == "[\n  1,\n  2,\n  3,\n  4\n]\n"


@pytest.mark.skipif(not cli.pyjq, reason="pyjq is not installed")
def test_paginate_json_jq(requests_mock):
    requests_mock.get(
        "https://example.com/",
        json={"rows": [1, 2]},
        headers={"link": '</?page=2>; rel="next"'},
    )
    requests_mock.get("https://example.com/?page=2", json={"rows": [3, 4]})
    result = CliRunner(mix_stderr=False).invoke(
        cli.cli, ["https://example.com/", "--jq", ".rows"]
    )
    assert result.exit_code == 0
    assert result.stdout == "[\n  1,\n  2,\n  3,\n  4\n]\n"


def test_header(requests_mock):
    requests_mock.get("https://example.com/", json=[1, 2])
    result = CliRunner(mix_stderr=False).invoke(
        cli.cli,
        ["https://example.com/", "--header", "foo", "bar", "--header", "baz", "1"],
    )
    assert result.exit_code == 0
    assert result.stdout == "[\n  1,\n  2\n]\n"
    sent_request = requests_mock.request_history[0]
    assert dict(sent_request.headers)["foo"] == "bar"
    assert dict(sent_request.headers)["baz"] == "1"


@pytest.mark.parametrize("ignore_errors", (False, True))
def test_ignore_http_errors(requests_mock, ignore_errors):
    requests_mock.get("https://example.com/", json=[1, 2], status_code=500)
    result = CliRunner(mix_stderr=False).invoke(
        cli.cli,
        ["https://example.com/"] + (["--ignore-http-errors"] if ignore_errors else []),
    )
    assert result.exit_code == 0 if ignore_errors else 1
    if ignore_errors:
        assert result.stdout == "[\n  1,\n  2\n]\n"
    else:
        assert result.stdout == ""
        assert "500 error fetching" in result.stderr



from setuptools import setup, find_packages
import io
import os

VERSION = "1.0"


def get_long_description():
    with io.open(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"),
        encoding="utf8",
    ) as fp:
        return fp.read()


setup(
    name="paginate-json",
    description="CLI tool for fetching paginated JSON from a URL",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Simon Willison",
    version=VERSION,
    license="Apache License, Version 2.0",
    packages=find_packages(),
    install_requires=["requests", "click"],
    extras_require={"test": ["pytest", "requests-mock", "cogapp"], "pyjq": ["pyjq"]},
    entry_points="""
        [console_scripts]
        paginate-json=paginate_json.cli:cli
    """,
    url="https://github.com/simonw/paginate-json",
    project_urls={
        "Changelog": "https://github.com/simonw/paginate-json/releases",
        "Source code": "https://github.com/simonw/paginate-json",
        "Issues": "https://github.com/simonw/paginate-json/issues",
        "CI": "https://github.com/simonw/paginate-json/actions",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)
